#!/usr/bin/env python3
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
"""JPEG XL conformance test runner.

Tool to perform a conformance test for a decoder.
"""

import argparse
import json
import numpy
import os
import shutil
import subprocess
import sys
import tempfile

import lcms2

BRIGHT_RED = "91"
BRIGHT_BLUE = "94"

def ColoredMessage(color, message):
    print(f"\033[{color}m{message}\033[0m", flush=True)

def Failure(message):
    ColoredMessage(BRIGHT_RED, message)
    return False

def CompareNPY(ref, ref_icc, dec, dec_icc, frame_idx, rmse_limit, peak_error):
    """Compare a decoded numpy against the reference one."""
    if ref.shape != dec.shape:
        return Failure(f'Expected shape {ref.shape} but found {dec.shape}')
    ref_frame = ref[frame_idx]
    dec_frame = dec[frame_idx]
    num_channels = ref_frame.shape[2]

    if ref_icc != dec_icc and peak_error > 0:
        # Transform colors before comparison.
        # Skip this if we expect fully lossless, since it introduces tiny errors
        # (even in the case where ref_icc and dec_icc are equivalent so it's a no-op)
        if num_channels < 3:
            return Failure(f"Only RGB images are supported")
        dec_clr = dec_frame[:, :, 0:3]
        dec_frame[:, :, 0:3] = lcms2.convert_pixels(dec_icc, ref_icc, dec_clr)

    error = numpy.abs(ref_frame - dec_frame)
    actual_peak_error = error.max()
    error_by_channel = [error[:, :, ch] for ch in range(num_channels)]
    actual_rmses = [numpy.sqrt(numpy.mean(error_ch * error_ch)) for error_ch in error_by_channel]
    actual_rmse = max(actual_rmses)

    print(f"RMSE: {actual_rmses}, actual peak: {actual_peak_error}", flush=True)

    if actual_rmse > rmse_limit:
        return Failure(f"RMSE too large: {actual_rmse} > {rmse_limit}")

    if actual_peak_error > peak_error:
        return Failure(
            f"Peak error too large: {actual_peak_error} > {peak_error}")
    return True


def CompareBinaries(ref_bin, dec_bin):
    """Compare a decoded binary file against the reference for exact contents."""
    with open(ref_bin, 'rb') as reff:
        ref_data = reff.read()

    with open(dec_bin, 'rb') as decf:
        dec_data = decf.read()

    if ref_data != dec_data:
        return Failure(
            f'Binary files mismatch: {ref_bin} {dec_bin}')
    return True


TEST_KEYS = set(
    ['reconstructed_jpeg', 'original_icc', 'rms_error', 'peak_error'])


def CheckMeta(dec, ref):
    if isinstance(ref, dict):
        if not isinstance(dec, dict):
            return Failure("Malformed metadata file")
        for k, v in ref.items():
            if k in TEST_KEYS:
                continue
            if k not in dec:
                return Failure(
                    f"Malformed metadata file: key {k} not found")
            vv = dec[k]
            return CheckMeta(vv, v)
    elif isinstance(ref, list):
        if not isinstance(dec, list) or len(dec) != len(ref):
            return Failure("Malformed metadata file")
        for vv, v in zip(dec, ref):
            return CheckMeta(vv, v)
    elif isinstance(ref, float):
        if not isinstance(dec, float):
            return Failure("Malformed metadata file")
        if abs(dec - ref) > 0.0001:
            return Failure(
                f"Metadata: Expected {ref}, found {dec}")
    elif dec != ref:
        return Failure(f"Metadata: Expected {ref}, found {dec}")
    return True


def RunTest(context, test_id, work_dir):
    ok = True
    test_dir = os.path.join(context.corpus_dir, test_id)

    with open(os.path.join(test_dir, 'test.json'), 'r') as f:
        descriptor = json.load(f)
        if 'sha256sums' in descriptor:
            del descriptor['sha256sums']

    exact_tests = []

    input_filename = os.path.join(test_dir, 'input.jxl')
    pixel_prefix = os.path.join(work_dir, 'decoded')
    output_filename = pixel_prefix + '_image.npy'
    cmd = context.decoder_cmd + [input_filename, output_filename]
    cmd_jpeg = []
    if 'preview' in descriptor:
        preview_filename = os.path.join(work_dir,
                                        'decoded_preview.npy')
        cmd.extend(['--preview_out', preview_filename])
    if 'reconstructed_jpeg' in descriptor:
        jpeg_filename = os.path.join(work_dir, 'reconstructed.jpg')
        cmd_jpeg = context.decoder_cmd + [input_filename, jpeg_filename]
        exact_tests.append(('reconstructed.jpg', jpeg_filename))
    if 'original_icc' in descriptor:
        decoded_original_icc = os.path.join(
            work_dir, 'decoded_org.icc')
        cmd.extend(['--orig_icc_out', decoded_original_icc])
        exact_tests.append(('original.icc', decoded_original_icc))
    meta_filename = os.path.join(work_dir, 'meta.json')
    cmd.extend(['--metadata_out', meta_filename])
    cmd.extend(['--icc_out', pixel_prefix + '.icc'])
    cmd.extend(['--norender_spotcolors'])

    print(f"Running: {cmd}", flush=True)
    if subprocess.call(cmd) != 0:
        return Failure('Running the decoder (%s) returned error' %
                        ' '.join(cmd))
    if cmd_jpeg:
        print(f"Running: {cmd_jpeg}", flush=True)
        if subprocess.call(cmd_jpeg) != 0:
            return Failure(
                'Running the decoder (%s) returned error' %
                ' '.join(cmd_jpeg))

    # Run validation of exact files.
    for reference_basename, decoded_filename in exact_tests:
        reference_filename = os.path.join(test_dir,
                                            reference_basename)
        binary_ok = CompareBinaries(reference_filename,
                                    decoded_filename)
        if not binary_ok and context.update_on_failure:
            os.unlink(reference_filename)
            shutil.copy2(decoded_filename, reference_filename)
            binary_ok = True
        ok = ok & binary_ok

    # Validate metadata.
    with open(meta_filename, 'r') as f:
        meta = json.load(f)

    ok = ok & CheckMeta(meta, descriptor)

    # Pixel data.
    decoded_icc = pixel_prefix + '.icc'
    with open(decoded_icc, 'rb') as f:
        decoded_icc = f.read()
    reference_icc = os.path.join(test_dir, "reference.icc")
    with open(reference_icc, 'rb') as f:
        reference_icc = f.read()

    reference_npy_fn = os.path.join(test_dir, 'reference_image.npy')
    decoded_npy_fn = os.path.join(work_dir, 'decoded_image.npy')

    if not os.path.exists(decoded_npy_fn):
        return Failure('File not decoded: decoded_image.npy')

    reference_npy = numpy.load(reference_npy_fn)
    decoded_npy = numpy.load(decoded_npy_fn)

    frames_ok = True
    for i, fd in enumerate(descriptor['frames']):
        frames_ok = frames_ok & CompareNPY(
            reference_npy, reference_icc, decoded_npy,
            decoded_icc, i, fd['rms_error'],
            fd['peak_error'])

    if not frames_ok and context.update_on_failure:
        os.unlink(reference_npy_fn)
        shutil.copy2(decoded_npy_fn, reference_npy_fn)
        frames_ok = True
    ok = ok & frames_ok

    if 'preview' in descriptor:
        reference_npy_fn = os.path.join(test_dir,
                                        'reference_preview.npy')
        decoded_npy_fn = os.path.join(work_dir,
                                        'decoded_preview.npy')

        if not os.path.exists(decoded_npy_fn):
            ok = Failure(
                'File not decoded: decoded_preview.npy')

        reference_npy = numpy.load(reference_npy_fn)
        decoded_npy = numpy.load(decoded_npy_fn)
        preview_ok = CompareNPY(reference_npy, reference_icc,
                                decoded_npy, decoded_icc, 0,
                                descriptor['preview']['rms_error'],
                                descriptor['preview']['peak_error'])
        if not preview_ok & args.update_on_failure:
            os.unlink(reference_npy_fn)
            shutil.copy2(decoded_npy_fn, reference_npy_fn)
            preview_ok = True
        ok = ok & preview_ok

    return ok

def ConformanceTestRunner(args):
    context = type('Expando', (object,), {})()
    context.update_on_failure = args.update_on_failure
    # We can pass either the .txt file or the directory which defaults to the
    # full corpus. This is useful to run a subset of the corpus in other .txt
    # files.
    if os.path.isdir(args.corpus):
        context.corpus_dir = args.corpus
        corpus_txt = os.path.join(args.corpus, 'corpus.txt')
    else:
        context.corpus_dir = os.path.dirname(args.corpus)
        corpus_txt = args.corpus
    context.decoder_cmd = args.decoder.strip().split(' ')

    all_tests = []
    failed_tests = []

    with open(corpus_txt, 'r') as f:
        for test_id in f:
            all_tests.append(test_id)
            test_id = test_id.rstrip('\n')
            ColoredMessage(BRIGHT_BLUE, f"Testing {test_id}")
            with tempfile.TemporaryDirectory(prefix=test_id) as work_dir:
                test_ok = RunTest(context, test_id, work_dir)
                if not test_ok:
                    failed_tests.append(test_id)

    all = len(all_tests)
    failed = len(failed_tests)
    passed = all - failed
    ColoredMessage(BRIGHT_BLUE, f"{passed} of {all} tests passed")
    if failed_tests:
        ColoredMessage(BRIGHT_RED, "Failed tests:")
        for test in failed_tests:
            ColoredMessage(BRIGHT_RED, f"  {test}")
        return False
    return True

def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--decoder',
                        metavar='DECODER',
                        required=True,
                        help='path to the decoder binary under test.')
    parser.add_argument(
        '--corpus',
        metavar='CORPUS',
        required=True,
        help=('path to the corpus directory or corpus descriptor'
              ' text file.'))
    parser.add_argument(
        '--update_on_failure', action='store_true',
        help='If set, updates reference files on failing checks.')
    args = parser.parse_args()
    if not ConformanceTestRunner(args):
        sys.exit(1)


if __name__ == '__main__':
    main()



# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Configuration file for the Sphinx documentation builder.
#
# See https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import re
import subprocess

def GetVersion():
    """Function to get the version of the current code."""
    with open(os.path.join(
            os.path.dirname(__file__), '../../lib/CMakeLists.txt'), 'r') as f:
        cmakevars = {}
        for line in f:
            m = re.match(r'set\(JPEGXL_([A-Z]+)_VERSION ([^\)]+)\)', line)
            if m:
                cmakevars[m.group(1)] = m.group(2)
    return '%s.%s.%s' % (cmakevars['MAJOR'], cmakevars['MINOR'], cmakevars['PATCH'])

def ConfigProject(app, config):
    # Configure the doxygen xml directory as the "xml" directory next to the
    # sphinx output directory. Doxygen generates by default the xml files in a
    # "xml" sub-directory of the OUTPUT_DIRECTORY.
    build_dir = os.path.dirname(app.outdir)
    xml_dir = os.path.join(build_dir, 'xml')
    config.breathe_projects['libjxl'] = xml_dir

    # Read the docs build environment doesn't run our cmake script so instead we
    # need to run doxygen manually here.
    if os.environ.get('READTHEDOCS', None) != 'True':
        return
    root_dir = os.path.realpath(os.path.join(app.srcdir, '../../'))
    doxyfile = os.path.join(build_dir, 'Doxyfile-rtd.doc')
    with open(doxyfile, 'w') as f:
        f.write(f"""
FILE_PATTERNS          = *.c *.h
GENERATE_HTML          = NO
GENERATE_LATEX         = NO
GENERATE_XML           = YES
INPUT                  = lib/include doc/api.txt
OUTPUT_DIRECTORY       = {build_dir}
PROJECT_NAME           = LIBJXL
QUIET                  = YES
RECURSIVE              = YES
STRIP_FROM_PATH        = lib/include
WARN_AS_ERROR          = YES
""")
    subprocess.check_call(['doxygen', doxyfile], cwd=root_dir)

def setup(app):
    # Generate doxygen XML on init when running from Read the docs.
    app.connect("config-inited", ConfigProject)

### Project information

project = 'libjxl'
project_copyright = 'JPEG XL Project Authors'
author = 'JPEG XL Project Authors'
version = GetVersion()

### General configuration

extensions = [
    # For integration with doxygen documentation.
    'breathe',
    # sphinx readthedocs theme.
    'sphinx_rtd_theme',
    # Do we use it?
    'sphinx.ext.graphviz',
]

breathe_default_project = 'libjxl'
breathe_projects = {}


# All the API is in C, except those files that end with cxx.h.
breathe_domain_by_extension = {'h': 'cpp'}
breathe_domain_by_file_pattern = {
    '*cxx.h': 'cpp',
}
breathe_implementation_filename_extensions = ['.cc']

# These are defined at build time by cmake.
c_id_attributes = [
    'JXL_EXPORT',
    'JXL_DEPRECATED',
    'JXL_THREADS_EXPORT',
]
cpp_id_attributes = c_id_attributes


breathe_projects_source = {
    'libjxl' : ('../../', [
        'doc/api.txt',
        'lib/include/jxl',
    ])
}

# Recognized suffixes.
source_suffix = ['.rst', '.md']

### Options for HTML output

# Use the readthedocs.io theme when generating the HTML output.
html_theme = 'sphinx_rtd_theme'



#!/usr/bin/env python3
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import shutil
import subprocess
import sys

from pathlib import Path

BROTLIFY = False
ZOPFLIFY = False
LEAN = True
NETLIFY = False

REMOVE_SHEBANG = ['jxl_decoder.js']
EMBED_BIN = [
  'jxl_decoder.js',
  'jxl_decoder.worker.js'
]
EMBED_SRC = ['client_worker.js']
TEMPLATES = ['service_worker.js']
COPY_BIN = ['jxl_decoder.wasm'] + [] if LEAN else EMBED_BIN
COPY_SRC = [
  'one_line_demo.html',
  'one_line_demo_with_console.html',
  'manual_decode_demo.html',
] + [] if not NETLIFY else [
  'netlify.toml',
  'netlify'
] + [] if LEAN else EMBED_SRC

COMPRESS = COPY_BIN + COPY_SRC + TEMPLATES
COMPRESSIBLE_EXT = ['.html', '.js', '.wasm']

def escape_js(js):
  return js.replace('\\', '\\\\').replace('\'', '\\\'')

def remove_shebang(txt):
  lines = txt.splitlines(True) # Keep line-breaks
  if len(lines) > 0:
    if lines[0].startswith('#!'):
      lines = lines[1:]
  return ''.join(lines)

def compress(path):
  name = path.name
  compressible = any([name.endswith(ext) for ext in COMPRESSIBLE_EXT])
  if not compressible:
    print(f'Not compressing {name}')
    return
  print(f'Processing {name}')
  orig_size = path.stat().st_size
  if BROTLIFY:
    cmd_brotli = ['brotli', '-Zfk', path.absolute()]
    subprocess.run(cmd_brotli, check=True, stdout=sys.stdout, stderr=sys.stderr)
    br_size = path.parent.joinpath(name + '.br').stat().st_size
    print(f'  Brotli: {orig_size} -> {br_size}')
  if ZOPFLIFY:
    cmd_zopfli = ['zopfli', path.absolute()]
    subprocess.run(cmd_zopfli, check=True, stdout=sys.stdout, stderr=sys.stderr)
    gz_size = path.parent.joinpath(name + '.gz').stat().st_size
    print(f'  Zopfli: {orig_size} -> {gz_size}')

def check_util(name):
  cmd = [name, '-h']
  try:
    subprocess.run(cmd, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
  except:
    print(f"NOTE: {name} not installed")
    return False
  return True

def check_utils():
  global BROTLIFY
  BROTLIFY = BROTLIFY and check_util('brotli')
  global ZOPFLIFY
  ZOPFLIFY = ZOPFLIFY and check_util('zopfli')
  if not check_util('uglifyjs'):
    print("FAIL: uglifyjs is required to build a site")
    sys.exit()

def uglify(text, name):
  cmd = ['uglifyjs', '-m', '-c']
  ugly_result = subprocess.run(
      cmd, capture_output=True, check=True, input=text, text=True)
  ugly_text = ugly_result.stdout.strip()
  print(f'Uglify {name}: {len(text)} -> {len(ugly_text)}')
  return ugly_text

if __name__ == "__main__":
  if len(sys.argv) != 4:
    print(f"Usage: python3 {sys.argv[0]} SRC_DIR BINARY_DIR OUTPUT_DIR")
    exit(-1)
  source_path = Path(sys.argv[1]) # CMake build dir
  binary_path = Path(sys.argv[2]) # Site template dir
  output_path = Path(sys.argv[3]) # Site output

  check_utils()

  for name in REMOVE_SHEBANG:
    path = binary_path.joinpath(name)
    text = path.read_text().strip()
    path.write_text(remove_shebang(text))
    remove_shebang

  substitutes = {}

  for name in EMBED_BIN:
    key = '$' + name + '$'
    path = binary_path.joinpath(name)
    value = escape_js(uglify(path.read_text().strip(), name))
    substitutes[key] = value

  for name in EMBED_SRC:
    key = '$' + name + '$'
    path = source_path.joinpath(name)
    value = escape_js(uglify(path.read_text().strip(), name))
    substitutes[key] = value

  for name in TEMPLATES:
    print(f'Processing template {name}')
    path = source_path.joinpath(name)
    text = path.read_text().strip()
    for key, value in substitutes.items():
      text = text.replace(key, value)
    #text = uglify(text, name)
    output_path.joinpath(name).write_text(text)

  for name in COPY_SRC:
    path = source_path.joinpath(name)
    if path.is_dir():
      shutil.copytree(path, output_path.joinpath(
          name).absolute(), dirs_exist_ok=True)
    else:
      shutil.copy(path, output_path.absolute())

  # TODO(eustas): uglify
  for name in COPY_BIN:
    shutil.copy(binary_path.joinpath(name), output_path.absolute())

  for name in COMPRESS:
    compress(output_path.joinpath(name))



#!/usr/bin/env python3
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import ctypes
from numpy.ctypeslib import ndpointer
import numpy
import os
import platform

IS_OSX = (platform.system() == "Darwin")

default_libcms2_lib_path = ["liblcms2.so.2", "liblcms2.2.dylib"][IS_OSX]
lcms2_lib_path = os.getenv("LCMS2_LIB_PATH", default_libcms2_lib_path)
lcms2_lib = ctypes.cdll.LoadLibrary(lcms2_lib_path)

native_open_profile = lcms2_lib.cmsOpenProfileFromMem
native_open_profile.restype = ctypes.c_void_p
native_open_profile.argtypes = [
    ctypes.c_char_p,  # MemPtr
    ctypes.c_size_t  # dwSize
]

native_close_profile = lcms2_lib.cmsCloseProfile
native_close_profile.restype = ctypes.c_int
native_close_profile.argtypes = [
    ctypes.c_void_p  # hProfile
]

native_create_transform = lcms2_lib.cmsCreateTransform
native_create_transform.restype = ctypes.c_void_p
native_create_transform.argtypes = [
    ctypes.c_void_p,  # Input
    ctypes.c_uint32,  # InputFormat
    ctypes.c_void_p,  # Output
    ctypes.c_uint32,  # OutputFormat
    ctypes.c_uint32,  # Intent
    ctypes.c_uint32  # dwFlags
]

native_delete_transform = lcms2_lib.cmsDeleteTransform
native_delete_transform.restype = None
native_delete_transform.argtypes = [
    ctypes.c_void_p  # hTransform
]

native_do_transform = lcms2_lib.cmsDoTransform
native_do_transform.restype = None
native_do_transform.argtypes = [
    ctypes.c_void_p,  # Transform
    ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),  # InputBuffer
    ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),  # OutputBuffer
    ctypes.c_uint32  # Size
]


def make_format(
    bytes_per_sample=4,  # float32
    num_channels=3,  # RGB or XYZ
    extra_channels=0,
    swap_channels=0,
    swap_endianness=0,
    planar=0,
    flavor=0,
    swap_first=0,
    unused=0,
    pixel_type=4,  # RGB
    optimized=0,
    floating_point=1):
    values = [bytes_per_sample, num_channels, extra_channels, swap_channels,
        swap_endianness, planar, flavor, swap_first, unused, pixel_type,
        optimized, floating_point]
    bit_width = [3, 4, 3, 1, 1, 1, 1, 1, 1, 5, 1, 1]
    result = 0
    shift = 0
    for i in range(len(bit_width)):
        result += values[i] << shift
        shift += bit_width[i]
    return result


def convert_pixels(from_icc, to_icc, from_pixels):
    from_icc = bytearray(from_icc)
    to_icc = bytearray(to_icc)

    if len(from_pixels.shape) != 3 or from_pixels.shape[2] != 3:
        raise ValueError("Only WxHx3 shapes are supported")
    from_pixels_plain = from_pixels.ravel().astype(numpy.float64)
    num_pixels = len(from_pixels_plain) // 3
    to_pixels_plain = numpy.empty(num_pixels * 3, dtype=numpy.float64)

    from_icc = (ctypes.c_char * len(from_icc)).from_buffer(from_icc)
    from_profile = native_open_profile(
        ctypes.cast(ctypes.pointer(from_icc), ctypes.c_char_p), len(from_icc))

    to_icc = (ctypes.c_char * len(to_icc)).from_buffer(to_icc)
    to_profile = native_open_profile(
        ctypes.cast(ctypes.pointer(to_icc), ctypes.c_char_p), len(to_icc))

    # bytes_per_sample=0 actually means 8 bytes (but there are just 3 bits to
    # encode the length of sample)
    format_rgb_f64 = make_format(bytes_per_sample=0)
    intent = 0  # INTENT_PERCEPTUAL
    flags = 0  # default; no "no-optimization"
    transform = native_create_transform(
        from_profile, format_rgb_f64, to_profile, format_rgb_f64, intent, flags)

    native_do_transform(
        transform, from_pixels_plain, to_pixels_plain, num_pixels)

    native_delete_transform(transform)
    native_close_profile(to_profile)
    native_close_profile(from_profile)

    # Return same shape and size as input
    return to_pixels_plain.reshape(from_pixels.shape).astype(from_pixels.dtype)

if __name__ == '__main__':
    raise Exception("Not an executable")



#!/usr/bin/env python3
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
"""Tool for generating a conformance testing corpus from a set of .jxl files.

This is not the JPEG XL conformance test runner. This is a tool to generate a
conformance testing corpus from a set of .jxl files.
"""

import argparse
import itertools
import json
import os
import shutil
import subprocess
import sys


def GenerateConformanceCorpus(args):
    """Generate the conformance test corpus for the given arguments."""
    files = []
    for jxl in args.inputs:
        if os.path.isdir(jxl):
            # Add all the .jxl files recursively.
            for root, _, dir_files in os.walk(jxl):
                files.extend(
                    os.path.join(root, filename) for filename in dir_files
                    if filename.lower().endswith('.jxl'))
        else:
            files.append(jxl)

    os.makedirs(args.output, 0o755, exist_ok=True)

    test_ids = []
    for jxl in files:
        # Generate a unique test_id for this file based on the filename.
        test_id = os.path.basename(jxl).lower()
        if test_id.endswith('.jxl'):
            test_id = test_id[:-4]
        if test_id in test_ids:
            for i in itertools.count(2):
                candidate = test_id + '%02d' % i
                if candidate not in test_ids:
                    test_id = candidate
                    break
        test_ids.append(test_id)

        test_dir = os.path.join(args.output, test_id)
        os.makedirs(test_dir, 0o755, exist_ok=True)
        print('Generating %s' % (test_id, ))
        input_file = os.path.join(test_dir, 'input.jxl')
        shutil.copy(jxl, input_file)

        # The test descriptor file.
        descriptor = {}
        descriptor['jxl'] = 'input.jxl'

        original_icc_filename = os.path.join(test_dir, 'original.icc')
        reconstructed_filename = os.path.join(test_dir, 'reconstructed.jpg')
        pixel_prefix = os.path.join(test_dir, 'reference')
        output_file = pixel_prefix + '_image.npy'
        cmd = [args.decoder, input_file, output_file]
        metadata_filename = os.path.join(test_dir, 'test.json')
        cmd.extend(['--metadata_out', metadata_filename])
        cmd.extend(['--icc_out', pixel_prefix + '.icc'])

        # Decode and generate the reference files.
        subprocess.check_call(cmd)

        with open(metadata_filename, 'r') as f:
            metadata = json.load(f)

        if os.path.exists(original_icc_filename):
            metadata['original_icc'] = "original.icc"

        if os.path.exists(reconstructed_filename):
            metadata['reconstructed_jpeg'] = "reconstructed.jpg"

        for frame in metadata['frames']:
            frame['rms_error'] = args.rmse
            frame['peak_error'] = args.peak_error

        if 'preview' in metadata:
            metadata['preview']['rms_error'] = args.rmse
            metadata['preview']['peak_error'] = args.peak_error

        # Create the test descriptor file.
        with open(metadata_filename, 'w') as f:
            json.dump(metadata, f, indent=2)

    # Generate a corpus descriptor with the list of the all the test_id names,
    # one per line.
    with open(os.path.join(args.output, 'corpus.txt'), 'w') as f:
        f.write(''.join(line + '\n' for line in test_ids))


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--decoder',
                        metavar='DECODER',
                        required=True,
                        help='path to the decoder binary under test.')
    parser.add_argument('--output',
                        metavar='DIR',
                        required=True,
                        help='path to the output directory')
    parser.add_argument('--peak_error',
                        metavar='PEAK_ERROR',
                        type=float,
                        required=True,
                        help='peak error for each testcase')
    parser.add_argument('--rmse',
                        metavar='RMSE',
                        type=float,
                        required=True,
                        help='max RMSE for each testcase')
    parser.add_argument('inputs',
                        metavar='JXL',
                        nargs='+',
                        help='path to input .jxl file(s)')
    args = parser.parse_args()
    GenerateConformanceCorpus(args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.


"""build_stats.py: Gather statistics about sizes of dependencies.

This tools computes a realistic estimate of the size contribution to a binary
from a statically linked library. Statically linked libraries compiled with
-ffunction-sections and linked -gc-sections mean that we could drop part of the
library at the final binary linking time. This tool takes that into account the
symbols that end up in the final binary and not just all the symbols of the
components.
"""

import argparse
import collections
import itertools
import json
import os
import platform
import re
import struct
import subprocess
import sys
import tempfile

# Ignore functions with stack size smaller than this value.
MIN_STACK_SIZE = 32

IS_OSX = (platform.system() == 'Darwin')

Symbol = collections.namedtuple('Symbol', ['address', 'size', 'kind', 'name'])

# Represents the stack size information of a function (defined by its address).
SymbolStack = collections.namedtuple('SymbolStack',
                                     ['address', 'stack_size'])

ObjectStats = collections.namedtuple('ObjectStats',
                                     ['name', 'in_partition', 'size_map'])

# An object target file in the build system.
Target = collections.namedtuple('Target',
                                ['name', 'deps', 'filename'])

# Sections that end up in the binary file.
# t - text (code), d - global non-const data, n/r - read-only data,
# w - weak symbols (likely inline code not inlined),
# v - weak symbols (vtable / typeinfo)
# u - unique symbols
BIN_SIZE = 'tdnrwvu'

# Sections that end up in static RAM.
RAM_SIZE = 'dbs'

# u - symbols imported from some other library
# a - absolute address symbols
# c - common symbol
# i - indirect symbol
# - - debugger symbol table entries
IGNORE_SYMBOLS = 'uaci-'

SIMD_NAMESPACES = [
    'N_SCALAR', 'N_EMU128',
    'N_RVV',
    'N_WASM', 'N_WASM_EMU256',
    'N_NEON', 'N_NEON_BF16', 'N_NEON_WITHOUT_AES',
    'N_SVE', 'N_SVE2', 'N_SVE_256', 'N_SVE2_128',
    'N_PPC8', 'N_PPC9', 'N_PPC10',
    'N_Z14', 'N_Z15',
    'N_SSE2', 'N_SSSE3', 'N_SSE4',
    'N_AVX2', 'N_AVX3', 'N_AVX3_DL', 'N_AVX3_ZEN4', 'N_AVX3_SPR'
]


def LoadSymbols(filename):
  ret = []
  nmout = subprocess.check_output(['nm', '--format=posix', filename])
  for line in nmout.decode('utf-8').splitlines():
    line = line.rstrip()
    if len(line) == 0:
      # OSX nm produces extra crlf at the end
      continue
    if line.endswith(':'):
      # Ignore object names.
      continue
    line = re.sub(' +', ' ', line)
    # symbol_name, symbol_type, (optional) address, (optional) size
    symlist = line.rstrip().split(' ')
    col_count = len(symlist)
    assert 2 <= col_count <= 4
    ret.append(Symbol(
        int(symlist[2], 16) if col_count > 2 else None,
        int(symlist[3], 16) if col_count > 3 else None,
        symlist[1],
        symlist[0]))
  if IS_OSX:
    ret = sorted(ret, key=lambda sym: sym.address)
    for i in range(len(ret) - 1):
      size = ret[i + 1].address - ret[i].address
      if size > (1 << 30):
        continue
      ret[i] = ret[i]._replace(size=size)
  return ret

def LoadTargetCommand(target, build_dir):
  stdout = subprocess.check_output(
      ['ninja', '-C', build_dir, '-t', 'commands', target])
  # The last command is always the command to build (link) the requested
  # target.
  command = stdout.splitlines()[-1]
  return command.decode('utf-8')


def LoadTarget(target, build_dir):
  """Loads a build system target and its dependencies into a Target object"""
  if target.endswith('.o'):
    # Speed up this case.
    return Target(target, [], target)

  link_params = LoadTargetCommand(target, build_dir).split()
  if 'cmake_symlink_library' in link_params:
    # The target is a library symlinked, use the target of the symlink
    # instead.
    target = link_params[link_params.index('cmake_symlink_library') + 1]
    link_params = LoadTargetCommand(target, build_dir).split()

  # The target name is not always the same as the filename of the output, for
  # example, "djxl" target generates "tools/djxl" file.
  if '-o' in link_params:
    target_filename = link_params[link_params.index('-o') + 1]
  elif target.endswith('.a'):
    # Command is '/path/to/ar', 'qc', 'target.a', ...
    target_filename = link_params[link_params.index('qc') + 1]
  else:
    raise Exception('Unknown "%s" output filename in command: %r' %
                    (target, link_params))

  tgt_libs = []
  for entry in link_params:
    if not entry or not (entry.endswith('.o') or entry.endswith('.a')):
      continue
    if entry == target_filename:
      continue
    fn = os.path.join(build_dir, entry)
    if not os.path.exists(fn):
      continue
    if entry in tgt_libs:
      continue
    tgt_libs.append(entry)

  return Target(target, tgt_libs, target_filename)


def TargetTransitiveDeps(all_tgts, target):
  """Returns the list of all transitive dependencies of target"""
  ret = all_tgts[target].deps
  # There can't be loop dependencies in the targets.
  i = 0
  while i < len(ret):
    ret.extend(all_tgts[ret[i]].deps)
    i += 1
  return ret


def LoadStackSizes(filename, binutils=''):
  """Loads the stack size used by functions from the ELF.

  This function loads the stack size the compiler stored in the .stack_sizes
  section, which can be done by compiling with -fstack-size-section in clang.
  """
  with tempfile.NamedTemporaryFile() as stack_sizes_sec:
    objcopy = ['objcopy', 'gobjcopy'][IS_OSX]
    subprocess.check_call(
        [binutils + objcopy, '-O', 'binary', '--only-section=.stack_sizes',
         '--set-section-flags', '.stack_sizes=alloc', filename,
         stack_sizes_sec.name])
    stack_sizes = stack_sizes_sec.read()
  # From the documentation:
  #  The section will contain an array of pairs of function symbol values
  #  (pointer size) and stack sizes (unsigned LEB128). The stack size values
  #  only include the space allocated in the function prologue. Functions with
  #  dynamic stack allocations are not included.

  # Get the pointer format based on the ELF file.
  objdump = ['objdump', 'gobjdump'][IS_OSX]
  output = subprocess.check_output(
      [binutils + objdump, '-a', filename]).decode('utf-8')
  elf_format = re.search('file format (.*)$', output, re.MULTILINE).group(1)
  if elf_format.startswith('elf64-little') or elf_format.endswith('-x86-64') or elf_format.endswith('-arm64'):
    pointer_fmt = '<Q'
  elif elf_format.startswith('elf32-little') or elf_format == 'elf32-i386':
    pointer_fmt = '<I'
  else:
    raise Exception('Unknown ELF format: %s' % elf_format)
  pointer_size = struct.calcsize(pointer_fmt)

  ret = []
  i = 0
  while i < len(stack_sizes):
    assert len(stack_sizes) >= i + pointer_size
    addr, = struct.unpack_from(pointer_fmt, stack_sizes, i)
    i += pointer_size
    # Parse LEB128
    size = 0
    for j in range(10):
      b = stack_sizes[i]
      i += 1
      size += (b & 0x7f) << (7 * j)
      if (b & 0x80) == 0:
        break
    if size >= MIN_STACK_SIZE:
      ret.append(SymbolStack(addr, size))
  return ret


def TargetSize(symbols, symbol_filter=None):
  ret = {}
  for sym in symbols:
    if not sym.size or (symbol_filter is not None and
                        sym.name not in symbol_filter):
      continue
    t = sym.kind.lower()
    # We can remove symbols if they appear in multiple objects since they will
    # be merged by the linker.
    if symbol_filter is not None and (t == sym.kind or t in 'wv'):
      symbol_filter.remove(sym.name)
    ret.setdefault(t, 0)
    ret[t] += sym.size
  return ret


def PrintStats(stats):
  """Print a table with the size stats for a target"""
  table = []
  sum_bin_size = 0
  sum_ram_size = 0

  for objstat in stats:
    bin_size = 0
    ram_size = 0
    for kind, size in objstat.size_map.items():
      if kind in BIN_SIZE:
        bin_size += size
      if kind in RAM_SIZE:
        ram_size += size
      if kind not in BIN_SIZE + RAM_SIZE:
        raise Exception('Unknown type "%s"' % kind)
    if objstat.in_partition:
      sum_bin_size += bin_size
      sum_ram_size += ram_size

    table.append((objstat.name, bin_size, ram_size))
  mx_bin_size = max(row[1] for row in table)
  mx_ram_size = max(row[2] for row in table)

  table.append(('-- unknown --', mx_bin_size - sum_bin_size,
                mx_ram_size - sum_ram_size))

  # Print the table
  print('%-32s %17s %17s' % ('Object name', 'Binary size', 'Static RAM size'))
  for name, bin_size, ram_size in table:
    print('%-32s %8d (%5.1f%%) %8d (%5.1f%%)' % (
        name, bin_size, (100. * bin_size / mx_bin_size) if mx_bin_size else 0,
        ram_size, (100. * ram_size / mx_ram_size) if mx_ram_size else 0))
  print()


def PrintStackStats(tgt_stack_sizes, top_entries=20):
  if not tgt_stack_sizes:
    return
  print(' Stack   Symbol name')
  for i, (name, size) in zip(itertools.count(), tgt_stack_sizes.items()):
    if top_entries > 0 and i >= top_entries:
      break
    print('%8d %s' % (size, name))
  print()


def PrintTopSymbols(tgt_top_symbols):
  if not tgt_top_symbols:
    return
  print(' Size     T Symbol name')
  for size, kind, name in tgt_top_symbols:
    print('%9d %s %s' % (size, kind, name))
  print()


def SizeStats(args):
  """Main entry point of the program after parsing parameters.

  Computes the size statistics of the given targets and their components."""
  # The dictionary with the stats that we store on disk as a json. This includes
  # one entry per passed args.target.
  stats = {}

  # Cache of Target object of a target.
  tgts = {}

  # Load all the targets.
  pending = set(args.target)
  while pending:
    target = pending.pop()
    tgt = LoadTarget(target, args.build_dir)
    tgts[target] = tgt
    if args.recursive:
      for dep in tgt.deps:
        if dep not in tgts:
          pending.add(dep)

  # Cache of symbols of a target.
  syms = {}
  # Load the symbols from the all targets and its deps.
  all_deps = set(tgts.keys()).union(*[set(tgt.deps) for tgt in tgts.values()])
  for entry in all_deps:
    fn = os.path.join(args.build_dir,
                      tgts[entry].filename if entry in tgts else entry)
    syms[entry] = LoadSymbols(fn)

  for target in args.target:
    tgt_stats = []
    tgt = tgts[target]

    tgt_syms = syms[target]
    used_syms = set()
    for sym in tgt_syms:
      if sym.kind.lower() in BIN_SIZE + RAM_SIZE:
        used_syms.add(sym.name)
      elif sym.kind.lower() in IGNORE_SYMBOLS:
        continue
      else:
        print('Unknown: %s %s' % (sym.kind, sym.name))

    target_path = os.path.join(args.build_dir, tgt.filename)
    sym_stacks = []
    if not target_path.endswith('.a'):
      sym_stacks = LoadStackSizes(target_path, args.binutils)
    symbols_by_addr = {sym.address: sym for sym in tgt_syms
                          if sym.kind.lower() in 'tw'}
    tgt_stack_sizes = collections.OrderedDict()
    for sym_stack in sorted(sym_stacks, key=lambda s: -s.stack_size):
      tgt_stack_sizes[
          symbols_by_addr[sym_stack.address].name] = sym_stack.stack_size

    tgt_top_symbols = []
    if args.top_symbols:
      tgt_top_symbols = [(sym.size, sym.kind, sym.name) for sym in tgt_syms
                         if sym.name in used_syms and sym.size]
      tgt_top_symbols.sort(key=lambda t: (-t[0], t[2]))
      tgt_top_symbols = tgt_top_symbols[:args.top_symbols]

    tgt_size = TargetSize(tgt_syms)
    tgt_stats.append(ObjectStats(target, False, tgt_size))

    # Split out by SIMD.
    for namespace in SIMD_NAMESPACES:
      mangled = str(len(namespace)) + namespace
      if not any(mangled in sym.name for sym in tgt_syms):
        continue
      ret = {}
      for sym in tgt_syms:
        if not sym.size or mangled not in sym.name:
          continue
        t = sym.kind.lower()
        ret.setdefault(t, 0)
        ret[t] += sym.size
      # SIMD namespaces are not part of the partition, they are already included
      # in the jpegxl-static normally.
      if not ret:
        continue
      tgt_stats.append(ObjectStats('\\--> ' + namespace, False, ret))

    for obj in tgt.deps:
      dep_used_syms = used_syms.copy()
      obj_size = TargetSize(syms[obj], used_syms)
      if not obj_size:
        continue
      tgt_stats.append(ObjectStats(os.path.basename(obj), True, obj_size))
      if args.recursive:
        # Not really recursive, but it shows all the remaining deps at a second
        # level.
        for obj_dep in sorted(TargetTransitiveDeps(tgts, obj),
                              key=os.path.basename):
          obj_dep_size = TargetSize(syms[obj_dep], dep_used_syms)
          if not obj_dep_size:
            continue
          tgt_stats.append(ObjectStats(
              '   '+ os.path.basename(obj_dep), False, obj_dep_size))

    PrintStats(tgt_stats)
    PrintStackStats(tgt_stack_sizes)
    PrintTopSymbols(tgt_top_symbols)
    stats[target] = {
        'build': tgt_stats,
        'stack': tgt_stack_sizes,
        'top': tgt_top_symbols,
    }

  if args.save:
    with open(args.save, 'w') as f:
      json.dump(stats, f)

  # Check the maximum stack size.
  exit_code = 0
  if args.max_stack:
    for name, size in tgt_stack_sizes.items():
      if size > args.max_stack:
        print('Error: %s exceeds stack limit: %d vs %d' % (
                  name, size, args.max_stack),
              file=sys.stderr)
        exit_code = 1

  return exit_code

def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument('target', type=str, nargs='+',
                      help='target(s) to analyze')
  parser.add_argument('--build-dir', default='build',
                      help='path to the build directory')
  parser.add_argument('--save', default=None,
                      help='path to save the stats as JSON file')
  parser.add_argument('-r', '--recursive', default=False, action='store_true',
                      help='Print recursive entries.')
  parser.add_argument('--top-symbols', default=0, type=int,
                      help='Number of largest symbols to print')
  parser.add_argument('--binutils', default='',
                      help='prefix path to binutils tools, such as '
                           'aarch64-linux-gnu-')
  parser.add_argument('--max-stack', default=None, type=int,
                      help=('Maximum static stack size of a function. If a '
                            'static stack is larger it will exit with an error '
                            'code.'))
  args = parser.parse_args()
  sys.exit(SizeStats(args))


if __name__ == '__main__':
  main()



#!/usr/bin/env python3
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.


"""build_cleaner.py: Update build files.

This tool keeps certain parts of the build files up to date.
"""

import argparse
import locale
import os
import re
import subprocess
import sys
import tempfile

COPYRIGHT = [
  "Copyright (c) the JPEG XL Project Authors. All rights reserved.",
  "",
  "Use of this source code is governed by a BSD-style",
  "license that can be found in the LICENSE file."
]

DOC = [
  "This file is generated, do not modify by manually.",
  "Run `tools/scripts/build_cleaner.py --update` to regenerate it."
]

def RepoFiles(src_dir):
  """Return the list of files from the source git repository"""
  git_bin = os.environ.get('GIT_BIN', 'git')
  files = subprocess.check_output([git_bin, '-C', src_dir, 'ls-files'])
  ret = files.decode(locale.getpreferredencoding()).splitlines()
  ret.sort()
  return ret


def Check(condition, msg):
  if not condition:
    print(msg)
    sys.exit(2)


def ContainsFn(*parts):
  return lambda path: any(part in path for part in parts)


def HasPrefixFn(*prefixes):
  return lambda path: any(path.startswith(prefix) for prefix in prefixes)


def HasSuffixFn(*suffixes):
  return lambda path: any(path.endswith(suffix) for suffix in suffixes)


def Filter(src, fn):
  yes_list = []
  no_list = []
  for item in src:
    (yes_list if fn(item) else no_list).append(item)
  return yes_list, no_list


def SplitLibFiles(repo_files):
  """Splits the library files into the different groups."""

  srcs_base = 'lib/'
  srcs, _ = Filter(repo_files, HasPrefixFn(srcs_base))
  srcs = [path[len(srcs_base):] for path in srcs]
  srcs, _ = Filter(srcs, HasSuffixFn('.cc', '.h', '.ui'))
  srcs.sort()

  # TODO(eustas): move to tools?
  _, srcs = Filter(srcs, HasSuffixFn('gbench_main.cc'))
  # This stub compilation unit is manually referenced in CMake buildfile.
  _, srcs = Filter(srcs, HasSuffixFn('nothing.cc'))

  # First pick files scattered across directories.
  tests, srcs = Filter(srcs, HasSuffixFn('_test.cc'))
  # TODO(eustas): move to separate list?
  _, srcs = Filter(srcs, ContainsFn('testing.h'))
  testlib_files, srcs = Filter(srcs, ContainsFn('test'))
  gbench_sources, srcs = Filter(srcs, HasSuffixFn('_gbench.cc'))

  extras_sources, srcs = Filter(srcs, HasPrefixFn('extras/'))
  lib_srcs, srcs = Filter(srcs, HasPrefixFn('jxl/'))
  public_headers, srcs = Filter(srcs, HasPrefixFn('include/jxl/'))
  threads_sources, srcs = Filter(srcs, HasPrefixFn('threads/'))

  Check(len(srcs) == 0, 'Orphan source files: ' + str(srcs))

  base_sources, lib_srcs = Filter(lib_srcs, HasPrefixFn('jxl/base/'))

  threads_public_headers, public_headers = Filter(
      public_headers, ContainsFn('_parallel_runner'))

  codec_names = ['apng', 'exr', 'gif', 'jpg', 'jxl', 'npy', 'pgx', 'pnm']
  codecs = {}
  for codec in codec_names:
    codec_sources, extras_sources = Filter(extras_sources, HasPrefixFn(
      f'extras/dec/{codec}', f'extras/enc/{codec}'))
    codecs[f'codec_{codec}_sources'] = codec_sources

  # TODO(eustas): move to separate folder?
  extras_for_tools_sources, extras_sources = Filter(extras_sources, ContainsFn(
    '/codec', '/hlg', '/metrics', '/packed_image_convert', '/render_hdr',
    '/tone_mapping'))

  # Source files only needed by the encoder or by tools (including decoding
  # tools), but not by the decoder library.
  # TODO(eustas): investigate the status of codec_in_out.h
  # TODO(eustas): rename butteraugli_wrapper.cc to butteraugli.cc?
  # TODO(eustas): is it possible to make butteraugli more standalone?
  enc_sources, lib_srcs = Filter(lib_srcs, ContainsFn('/enc_', '/butteraugli',
    'jxl/encode.cc', 'jxl/encode_internal.h'
  ))

  # The remaining of the files are in the dec_library.
  dec_jpeg_sources, dec_sources = Filter(lib_srcs, HasPrefixFn('jxl/jpeg/',
    'jxl/decode_to_jpeg.cc', 'jxl/decode_to_jpeg.h'))
  dec_box_sources, dec_sources = Filter(dec_sources, HasPrefixFn(
    'jxl/box_content_decoder.cc', 'jxl/box_content_decoder.h'))
  cms_sources, dec_sources = Filter(dec_sources, HasPrefixFn('jxl/cms/'))

  # TODO(lode): further prune dec_srcs: only those files that the decoder
  # absolutely needs, and or not only for encoding, should be listed here.

  return codecs | {'base_sources': base_sources,
    'cms_sources': cms_sources,
    'dec_box_sources': dec_box_sources,
    'dec_jpeg_sources': dec_jpeg_sources,
    'dec_sources': dec_sources,
    'enc_sources': enc_sources,
    'extras_for_tools_sources': extras_for_tools_sources,
    'extras_sources': extras_sources,
    'gbench_sources': gbench_sources,
    'public_headers': public_headers,
    'testlib_files': testlib_files,
    'tests': tests,
    'threads_public_headers': threads_public_headers,
    'threads_sources': threads_sources,
  }


def MaybeUpdateFile(args, filename, new_text):
  """Optionally replace file with new contents.

  If args.update is set, it will update the file with the new contents,
  otherwise it will return True when no changes were needed.
  """
  filepath = os.path.join(args.src_dir, filename)
  with open(filepath, 'r') as f:
    src_text = f.read()

  if new_text == src_text:
    return True

  if args.update:
    print('Updating %s' % filename)
    with open(filepath, 'w') as f:
      f.write(new_text)
    return True
  else:
    prefix = os.path.basename(filename)
    with tempfile.NamedTemporaryFile(mode='w', prefix=prefix) as new_file:
      new_file.write(new_text)
      new_file.flush()
      subprocess.call(['diff', '-u', filepath, '--label', 'a/' + filename,
        new_file.name, '--label', 'b/' + filename])
    return False


def FormatList(items, prefix, suffix):
  return ''.join(f'{prefix}{item}{suffix}\n' for item in items)


def FormatGniVar(name, var):
  if type(var) is list:
    contents = FormatList(var, '    "', '",')
    return f'{name} = [\n{contents}]\n'
  else:  # TODO(eustas): do we need scalar strings?
    return f'{name} = {var}\n'


def FormatCMakeVar(name, var):
  if type(var) is list:
    contents = FormatList(var, '  ', '')
    return f'set({name}\n{contents})\n'
  else:  # TODO(eustas): do we need scalar strings?
    return f'set({name} {var})\n'


def ToHashComment(lines):
  return [("# " + line).rstrip() for line in lines]
def ToDocstringComment(lines):
  return ["\"\"\""] + lines + ["\"\"\""]

def BuildCleaner(args):
  repo_files = RepoFiles(args.src_dir)

  with open(os.path.join(args.src_dir, 'lib/CMakeLists.txt'), 'r') as f:
    cmake_text = f.read()
  version = {'major_version': '', 'minor_version': '', 'patch_version': ''}
  for var in version.keys():
    cmake_var = f'JPEGXL_{var.upper()}'
    # TODO(eustas): use `cmake -L`
    # Regexp:
    #   set(_varname_ _capture_decimal_)
    match = re.search(r'set\(' + cmake_var + r' ([0-9]+)\)', cmake_text)
    version[var] = match.group(1)

  lists = SplitLibFiles(repo_files)

  cmake_chunks = ToHashComment(COPYRIGHT) + [""] + ToHashComment(DOC)
  cmake_parts = lists
  for var in sorted(cmake_parts):
    cmake_chunks.append(FormatCMakeVar(
        'JPEGXL_INTERNAL_' + var.upper(), cmake_parts[var]))

  gni_bzl_parts = version | lists
  gni_bzl_chunks = []
  for var in sorted(gni_bzl_parts):
    gni_bzl_chunks.append(FormatGniVar('libjxl_' + var, gni_bzl_parts[var]))

  bzl_chunks = ToHashComment(COPYRIGHT) + [""] + \
      ToDocstringComment(DOC) + [""] + gni_bzl_chunks
  gni_chunks = ToHashComment(COPYRIGHT) + [""] + \
      ToHashComment(DOC) + [""] + gni_bzl_chunks

  okay = [
    MaybeUpdateFile(args, 'lib/jxl_lists.bzl', '\n'.join(bzl_chunks)),
    MaybeUpdateFile(args, 'lib/jxl_lists.cmake', '\n'.join(cmake_chunks)),
    MaybeUpdateFile(args, 'lib/lib.gni', '\n'.join(gni_chunks)),
  ]
  return all(okay)


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument('--src-dir',
    default=os.path.realpath(os.path.join( os.path.dirname(__file__), '../..')),
    help='path to the build directory')
  parser.add_argument('--update', default=False, action='store_true',
    help='update the build files instead of only checking')
  args = parser.parse_args()
  Check(BuildCleaner(args), 'Build files need update.')


if __name__ == '__main__':
  main()



#!/usr/bin/env python3

# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""Generates coefficients used in upscaling.

Given an upscaling factor which can be 2, 4 or 8, we generate coefficients and
indices for lib/jxl/image_metadata.cc in the format needed there.
"""

import argparse
import itertools
import numpy as np


def compute_kernel(sigma):
  """Gaussian-like kernel with standard deviation sigma."""
  # This controls the length of the kernel.
  m = 2.5
  diff = int(max(1, m * abs(sigma)))
  kernel = np.exp(-np.arange(-diff, diff + 1)**2 /(2 * sigma * sigma))
  return kernel


def convolution(pixels, kernel):
  """Computes a horizontal convolution and transposes the result."""
  y, x = pixels.shape
  kernel_len = len(kernel)
  offset = kernel_len // 2
  scale = 1 / sum(kernel)
  out_pixels = np.zeros(shape=(x, y), dtype=pixels.dtype)
  for i, j in itertools.product(range(x), range(y)):
    if kernel_len < i < x - kernel_len:
      out_pixels[i, j] = scale * sum(
          pixels[j, i - offset + k] * kernel[k] for k in range(kernel_len))
    else:
      out_pixels[i, j] = pixels[j, i]
  return out_pixels


def _super_sample(pixels, n):
  return np.repeat(np.repeat(pixels, n, axis=0), n, axis=1)


def _sub_sample(pixels, n):
  x, y = pixels.shape
  assert x%n == 0 and y%n == 0
  return 1 / (n * n) * pixels.reshape(x // n, n, y // n, n).transpose(
      [0, 2, 1, 3]).sum(axis=(2, 3))


def smooth_4x4_corners(pixels):
  """Generates a 4x4 upscaled image, to be smoothed afterwards."""
  overshoot = 3.5
  m = 1.0 / (4.0 - overshoot)
  y_size, x_size = pixels.shape
  for y, x in itertools.product(range(3, y_size - 3, 4),
                                range(3, x_size - 3, 4)):
    ave = (
        pixels[y, x] + pixels[y, x + 1] + pixels[y + 1, x] +
        pixels[y + 1, x + 1])
    off = 2
    other = (ave - overshoot * pixels[y, x]) * m
    pixels[y - off, x - off] -= (other - pixels[y, x])
    pixels[y, x] = other

    other = (ave - overshoot * pixels[y, x + 1]) * m
    pixels[y - off, x + off + 1] -= (other - pixels[y, x + 1])
    pixels[y, x + 1] = other

    other = (ave - overshoot * pixels[y + 1, x]) * m
    pixels[y + off + 1, x - off] -= (other - pixels[y + 1, x])
    pixels[y + 1, x] = other

    other = (ave - overshoot * pixels[y + 1, x + 1]) * m
    pixels[y + off + 1][x + off + 1] -= (other - pixels[y + 1, x + 1])
    pixels[y + 1, x + 1] = other

  return pixels


def smoothing(pixels):
  new_pixels = smooth_4x4_corners(_super_sample(pixels, 4))
  my_kernel = compute_kernel(2.5)
  smooth_image = convolution(convolution(new_pixels, my_kernel), my_kernel)
  return smooth_image


upscaling = {
    2: lambda pixels: _sub_sample(smoothing(pixels), 2),
    4: smoothing,
    8: lambda pixels: _sub_sample(smoothing(smoothing(pixels)), 2)
}


def get_coeffs(upscaling_factor, kernel_size=5, normalized=True, dtype="float"):
  """Returns 4-tensor of coefficients.

  Args:
    upscaling_factor: 2, 4, or 8
    kernel_size: must be odd
    normalized: if True, the kernel matrix adds to 1
    dtype: type of numpy array to return

  Returns:
    A (upscaling_factor x upscaling_factor) matrix of
    (kernel_size x kernel_size) matrices, describing the kernel for all pixels.
  """

  upscaling_method = upscaling[upscaling_factor]
  patch_size = 2 * kernel_size + 1
  matrix_bases = np.eye(
      patch_size * patch_size, dtype=dtype).reshape(patch_size, patch_size,
                                                    patch_size, patch_size)

  # takes some time...
  smoothed_bases = np.array(
      [[upscaling_method(matrix_bases[a, b])
        for a in range(patch_size)]
       for b in range(patch_size)])

  middle = patch_size // 2
  lower = middle - kernel_size // 2
  upper = middle + kernel_size // 2 + 1
  assert len(range(lower, upper)) == kernel_size
  assert sum(range(lower, upper)) == kernel_size * middle

  coefficients = np.array([[[[
      smoothed_bases[i, j, upscaling_factor * middle + b,
                     upscaling_factor * middle + a]
      for i in range(lower, upper)
  ]
                             for j in range(lower, upper)]
                            for a in range(upscaling_factor)]
                           for b in range(upscaling_factor)])

  if normalized:
    return coefficients / coefficients.sum(axis=(2, 3))[..., np.newaxis,
                                                        np.newaxis]
  else:
    return coefficients


def indices_matrix(upscaling_factor, kernel_size=5):
  """Matrix containing indices with all symmetries."""
  matrix = np.zeros(
      shape=[upscaling_factor * kernel_size] * 2, dtype="int16")
  # define a fundamental domain
  counter = 1
  for i in range((kernel_size * upscaling_factor) // 2):
    for j in range(i, (kernel_size * upscaling_factor) // 2):
      matrix[i, j] = counter
      counter += 1

  matrix_with_transpose = matrix + (matrix.transpose()) * (
      matrix != matrix.transpose())
  matrix_vertical = matrix_with_transpose + (
      np.flip(matrix_with_transpose, axis=0) *
      (matrix_with_transpose != np.flip(matrix_with_transpose, axis=0)))
  matrix_horizontal = matrix_vertical + (
      np.flip(matrix_vertical, axis=1) *
      (matrix_vertical != np.flip(matrix_vertical, axis=1))) - 1
  return matrix_horizontal


def format_indices_matrix(upscaling_factor, kernel_size=5):
  """Returns string of commented out numbers-only matrices."""
  indices = indices_matrix(upscaling_factor)
  output_str = []
  for i in range(upscaling_factor // 2):
    for j in range(kernel_size):
      output_str.append("//")
      for a in range(upscaling_factor // 2):
        for b in range(kernel_size):
          output_str.append(
              f"{'{:x}'.format(int(indices[kernel_size*i + j][kernel_size*a + b])).rjust(2)} "
          )
        output_str.append(" ")
      output_str.append("\n")
    output_str.append("\n")
  return "".join(output_str)


def weights_arrays(upscaling_factor, kernel_size=5):
  """Returns string describing array of depth 4."""
  indices = indices_matrix(upscaling_factor)
  return (
      f"kernel[{upscaling_factor}][{upscaling_factor}][{kernel_size}][{kernel_size}]"
      f" = {{" + ", \n".join("{\n" + ", \n\n".join(
          ("{" + ", \n".join("{" + ", ".join(
              f"weights[{str(indices[kernel_size*i + j][kernel_size*a + b])}]"
              for b in range(kernel_size)) + "}"
                             for j in range(kernel_size)) + "}"
           for a in range(upscaling_factor // 2))) + "\n}"
                             for i in range(upscaling_factor // 2)) + "}\n")


def coefficients_list(upscaling_factor, kernel_size=5):
  """Returns string describing coefficients."""
  coeff_tensor = get_coeffs(upscaling_factor,
                            kernel_size).transpose([0, 2, 1, 3]).reshape(
                                kernel_size * upscaling_factor,
                                kernel_size * upscaling_factor)
  my_weights = [
      f'{"{:.8f}".format(coeff_tensor[i][j])}f'
      for i in range((kernel_size * upscaling_factor) // 2)
      for j in range(i, (kernel_size * upscaling_factor) // 2)
  ]
  return f"kWeights{upscaling_factor} = {{" + ", ".join(my_weights) + "};"


def print_all_output(upscaling_factor):
  print(format_indices_matrix(upscaling_factor))
  print(coefficients_list(upscaling_factor), end="\n\n")
  print(weights_arrays(upscaling_factor))


def main():
  parser = argparse.ArgumentParser(
      description="Generates coefficients used in upscaling.")
  parser.add_argument(
      "upscaling_factor",
      type=int,
      help="upscaling factor, must be  2, 4 or 8.",
      nargs="?",
      default=None)

  args = parser.parse_args()
  upscaling_factor = args.upscaling_factor
  if upscaling_factor:
    print_all_output(upscaling_factor)
  else:
    for factor in [2, 4, 8]:
      print(f"upscaling factor = {factor}")
      print_all_output(factor)


if __name__ == "__main__":
  main()



#!/usr/bin/env python3
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import csv
import sys
import math
import plotly.graph_objects as go

_, results, output_dir, *rest = sys.argv
OUTPUT = rest[0] if rest else 'svg'
# valid values: html, svg, png, webp, jpeg, pdf

with open(results, 'r') as f:
    reader = csv.DictReader(f)
    all_results = list(reader)

nonmetric_columns = set([
    "method", "image", "error", "size", "pixels", "enc_speed", "dec_speed",
    "bpp", "bppp", "qabpp"
])

metrics = set(all_results[0].keys()) - nonmetric_columns


def codec(method):
    sm = method.split(':')
    ssm = set(sm)
    speeds = set([
        'kitten', 'falcon', 'wombat', 'cheetah', 'tortoise', 'squirrel',
        'hare', 'fast'
    ])
    s = speeds.intersection(ssm)
    if sm[0] == 'custom':
        return sm[1]
    if sm[0] == 'jxl' and s:
        return 'jxl-' + list(s)[0]
    return sm[0]


data = {(m, img): {c: []
                   for c in {codec(x['method'])
                             for x in all_results}}
        for m in metrics for img in {x['image']
                                     for x in all_results}}

for r in all_results:
    c = codec(r['method'])
    img = r['image']
    bpp = r['bpp']
    for m in metrics:
        data[(m, img)][c].append((float(bpp), float(r[m])))


def pos(codec):
    if 'jxl-dis' in codec:
        return 6, codec
    elif 'jxl' in codec:
        return 7, codec
    elif 'avif' in codec:
        return 5, codec
    elif 'kdu' in codec:
        return 4, codec
    elif 'heif' in codec:
        return 3, codec
    elif 'fuif' in codec or 'pik' in codec:
        return 2, codec
    elif 'jpg' in codec or 'jpeg' in codec or 'web' in codec:
        return 1, codec
    else:
        return 0, codec


def style(codec):
    configs = {
        'jxl-cheetah': {
            'color': '#e41a1c',
            'dash': '1px, 1px',
            'width': 2
        },
        'jxl-wombat': {
            'color': '#e41a1c',
            'dash': '2px, 2px',
            'width': 2
        },
        'jxl-squirrel': {
            'color': '#e41a1c',
            'dash': '5px, 5px',
            'width': 2
        },
        'jxl-kitten': {
            'color': '#e41a1c',
            'width': 2
        },
        'jxl-dis-cheetah': {
            'color': '#377eb8',
            'dash': '1px, 1px',
            'width': 2
        },
        'jxl-dis-wombat': {
            'color': '#377eb8',
            'dash': '2px, 2px',
            'width': 2
        },
        'jxl-dis-squirrel': {
            'color': '#377eb8',
            'dash': '5px, 5px',
            'width': 2
        },
        'jxl-dis-kitten': {
            'color': '#377eb8',
            'width': 2
        },
        'rav1e.avif': {
            'color': '#4daf4a',
            'dash': '3px, 3px',
            'width': 2
        },
        '420.rav1e.avif': {
            'color': '#4daf4a',
            'dash': '1px, 1px',
            'width': 2
        },
        '444.rav1e.avif': {
            'color': '#4daf4a',
            'dash': '3px, 3px',
            'width': 2
        },
        'psnr.420.aom.avif': {
            'color': '#4daf4a',
            'dash': '5px, 5px',
            'width': 2
        },
        'psnr.444.aom.avif': {
            'color': '#4daf4a',
            'dash': '7px, 7px',
            'width': 2
        },
        'ssim.420.aom.avif': {
            'color': '#4daf4a',
            'dash': '9px, 9px',
            'width': 2
        },
        'ssim.444.aom.avif': {
            'color': '#4daf4a',
            'width': 2
        },
        'heif': {
            'color': '#984ea3',
            'width': 2
        },
        'fuif': {
            'color': '#ff7f00',
            'dash': '2px, 2px',
            'width': 2
        },
        'pik-cfp': {
            'color': '#ff7f00',
            'width': 2
        },
        'pik-cfp-fast': {
            'color': '#ff7f00',
            'dash': '4px, 4px',
            'width': 2
        },
        'webp': {
            'color': '#000000',
            'width': 2
        },
        'jpeg': {
            'color': '#a65628',
            'width': 2
        },
        'xt.jpg': {
            'color': '#a65628',
            'width': 2
        },
        'perc1.kdu.j2k': {
            'color': '#f781bf',
            'dash': '1px, 1px',
            'width': 2
        },
        'perc2.kdu.j2k': {
            'color': '#f781bf',
            'dash': '3px, 3px',
            'width': 2
        },
        'perc3.kdu.j2k': {
            'color': '#f781bf',
            'dash': '5px, 5px',
            'width': 2
        },
        'perc4.kdu.j2k': {
            'color': '#f781bf',
            'dash': '7px, 7px',
            'width': 2
        },
        'default.kdu.j2k': {
            'color': '#f781bf',
            'width': 2
        },
    }
    return configs.get(codec, dict())


visible_by_default = set([
    'jxl-kitten', 'ssim.444.aom.avif', 'heif', 'webp', 'jpeg', 'xt.jpg',
    'default.kdu.j2k'
])

column_remap = {
    'p': '6-Butteraugli',
    'dist': 'Max-Butteraugli',
    'psnr': "PSNR-YUV 6/8 Y",
    'MS-SSIM-Y': '-log10(1 - MS-SSIM-Y)',
    'puSSIM': '-log10(1 - puSSIM)',
    'FSIM-Y': '-log10(1 - FSIM-Y)',
    'FSIM-RGB': '-log10(1 - FSIM-RGB)',
    'VMAF': '-log10(1 - VMAF / 100)',
}


def remap(metric):
    funs = {
        'MS-SSIM-Y': lambda x: -math.log10(1 - x),
        'puSSIM': lambda x: -math.log10(1 - x),
        'FSIM-Y': lambda x: -math.log10(1 - x),
        'FSIM-RGB': lambda x: -math.log10(1 - x),
        'VMAF': lambda x: -math.log10(1 + 1e-8 - x / 100),
    }
    return funs.get(metric, lambda x: x)


for (m, img) in data:
    fname = "%s/%s_%s" % (output_dir, m, img)
    fig = go.Figure()
    for method in sorted(data[(m, img)].keys(), key=pos):
        vals = data[(m, img)][method]
        zvals = list(zip(*sorted(vals)))
        if not zvals:
            continue
        fig.add_trace(
            go.Scatter(x=zvals[0],
                       y=[remap(m)(x) for x in zvals[1]],
                       mode='lines',
                       name=method,
                       line=style(method),
                       visible=True
                       if method in visible_by_default else 'legendonly'))
    fig.update_layout(title=img,
                      xaxis_title='bpp',
                      yaxis_title=column_remap.get(m, m))
    fig.update_xaxes(type='log')
    if OUTPUT == 'html':
        fig.write_html(fname + '.html', include_plotlyjs='directory')
    else:
        fig.write_image(fname + '.' + OUTPUT, scale=4)



#!/usr/bin/python
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""Implementation of simplex search for an external process.

The external process gets the input vector through environment variables.
Input of vector as setenv("VAR%dimension", val)
Getting the optimized function with regexp match from stdout
of the forked process.

https://en.wikipedia.org/wiki/Nelder%E2%80%93Mead_method

start as ./simplex_fork.py binary dimensions amount
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from six.moves import range
import copy
import os
import random
import re
import subprocess
import sys

def Midpoint(simplex):
  """Nelder-Mead-like simplex midpoint calculation."""
  simplex.sort()
  dim = len(simplex) - 1
  retval = [None] + [0.0] * dim
  for i in range(1, dim + 1):
    for k in range(dim):
      retval[i] += simplex[k][i]
    retval[i] /= dim
  return retval


def Subtract(a, b):
  """Vector arithmetic, with [0] being ignored."""
  return [None if k == 0 else a[k] - b[k] for k in range(len(a))]

def Add(a, b):
  """Vector arithmetic, with [0] being ignored."""
  return [None if k == 0 else a[k] + b[k] for k in range(len(a))]

def Average(a, b):
  """Vector arithmetic, with [0] being ignored."""
  return [None if k == 0 else 0.5 * (a[k] + b[k]) for k in range(len(a))]


eval_hash = {}
g_best_val = None

def EvalCacheForget():
  global eval_hash
  eval_hash = {}

g_first_pnorm = None

def RandomizedJxlCodecs():
  global g_first_pnorm
  g_first_pnorm = None
  retval = []
  minval = 1.0
  maxval = 8.3
  rangeval = maxval/minval
  steps = 5
  for i in range(steps):
    mul = minval * rangeval**(float(i)/(steps - 1))
    mul *= 0.99 + 0.05 * random.random()
    retval.append("jxl:d%.4f" % mul)
  for i in range(steps - 1):
    mul = minval * rangeval**(float(i+0.5)/(steps - 1))
    mul *= 0.99 + 0.05 * random.random()
    retval.append("jxl:d%.4f" % mul)
  return ",".join(retval)

g_codecs = RandomizedJxlCodecs()

def Eval(vec, binary_name, cached=True):
  """Evaluates the objective function by forking a process.

  Args:
    vec: [0] will be set to the objective function, [1:] will
      contain the vector position for the objective function.
    binary_name: the name of the binary that evaluates the value.
  """
  global eval_hash
  global g_codecs
  global g_best_val
  global g_first_pnorm
  key = ""
  # os.environ["BUTTERAUGLI_OPTIMIZE"] = "1"
  for i in range(300):
    os.environ["VAR%d" % i] = "0"
  for i in range(len(vec) - 1):
    os.environ["VAR%d" % i] = str(vec[i + 1])
    key += str(vec[i + 1]) + ":"
  if cached and (key in eval_hash):
    vec[0] = eval_hash[key]
    return

  process = subprocess.Popen(
      (binary_name,
       '--input',
       '/usr/local/google/home/jyrki/newcorpus/split/*.png',
       '--error_pnorm=3.0',
       '--more_columns',
       '--codec', g_codecs),
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE,
      env=dict(os.environ))

  # process.wait()
  found_score = False
  vec[0] = 1e29
  dct2 = 0.0
  dct4 = 0.0
  dct16 = 0.0
  dct32 = 0.0
  n = 0
  for line in process.communicate(input=None)[0].splitlines():
    print("BE", line)
    sys.stdout.flush()
    if line[0:4] == b'Aggr':
      bpppnorm = float(line.split()[10])
      if g_first_pnorm == None:
        g_first_pnorm = float(line.split()[9])
      correctbpp = 1. + 10.0 * (float(line.split()[9]) - g_first_pnorm) ** 2

      vec[0] = 1e-5 * bpppnorm * correctbpp
      found_score = True

  print("eval: ", vec)
  if (vec[0] <= 0.0):
    vec[0] = 1e30
  if found_score:
    eval_hash[key] = vec[0]
    if not g_best_val or vec[0] < g_best_val:
      g_best_val = vec[0]
      print("\nSaving best simplex\n")
      with open("best_simplex.txt", "w") as f:
        print(vec, file=f)
    return
  vec[0] = 1e33
  return
  # sys.exit("awful things happened")

def Reflect(simplex, binary):
  """Main iteration step of Nelder-Mead optimization. Modifies `simplex`."""
  simplex.sort()
  last = simplex[-1]
  mid = Midpoint(simplex)
  diff = Subtract(mid, last)
  mirrored = Add(mid, diff)
  Eval(mirrored, binary)
  if mirrored[0] > simplex[-2][0]:
    print("\nStill worst\n\n")
    # Still the worst, shrink towards the best.
    shrinking = Average(simplex[-1], simplex[0])
    Eval(shrinking, binary)
    print("\nshrinking...\n\n")
    simplex[-1] = shrinking
    return
  if mirrored[0] < simplex[0][0]:
    # new best
    print("\nNew Best\n\n")
    even_further = Add(mirrored, diff)
    Eval(even_further, binary)
    if even_further[0] < mirrored[0]:
      print("\nEven Further\n\n")
      mirrored = even_further
    simplex[-1] = mirrored
    # try to extend
    return
  else:
    # not a best, not a worst point
    simplex[-1] = mirrored


def OneDimensionalSearch(simplex, shrink, index):
  # last appended was better than the best so far, try to replace it
  last_attempt = simplex[-1][:]
  best = simplex[0]
  if last_attempt[0] < best[0]:
    # try expansion of the amount
    diff = simplex[-1][index] - simplex[0][index]
    simplex[-1][index] = simplex[0][index] + shrink * diff
    Eval(simplex[-1], g_binary)
    if simplex[-1][0] < last_attempt[0]:
      # it got better
      return True
  elif last_attempt[0] >= 0:
    diff = simplex[-1][index] - simplex[0][index]
    simplex[-1][index] = simplex[0][index] - diff
    Eval(simplex[-1], g_binary)
    if simplex[-1][0] < last_attempt[0]:
      # it got better
      return True
  simplex[-1] = last_attempt
  return False

def InitialSimplex(vec, dim, amount):
  """Initialize the simplex at origin."""
  EvalCacheForget()
  best = vec[:]
  Eval(best, g_binary)
  retval = [best]
  comp_order = list(range(1, dim + 1))
  random.shuffle(comp_order)

  for i in range(dim):
    index = comp_order[i]
    best = retval[0][:]
    best[index] += amount
    Eval(best, g_binary)
    retval.append(best)
    do_shrink = True
    while OneDimensionalSearch(retval, 2.0, index):
      print("OneDimensionalSearch-Grow")
    while OneDimensionalSearch(retval, 1.1, index):
      print("OneDimensionalSearch-SlowGrow")
      do_shrink = False
    if do_shrink:
      while OneDimensionalSearch(retval, 0.9, index):
        print("OneDimensionalSearch-SlowShrinking")
    retval.sort()
  return retval


if len(sys.argv) != 4:
  print("usage: ", sys.argv[0], "binary-name number-of-dimensions simplex-size")
  exit(1)

g_dim = int(sys.argv[2])
g_amount = float(sys.argv[3])
g_binary = sys.argv[1]
g_simplex = InitialSimplex([None] + [0.0] * g_dim,
                           g_dim, 7.0 * g_amount)
best = g_simplex[0][:]
g_codecs = RandomizedJxlCodecs()
g_simplex = InitialSimplex(best, g_dim, g_amount * 2.47)
best = g_simplex[0][:]
g_simplex = InitialSimplex(best, g_dim, g_amount)
best = g_simplex[0][:]
g_simplex = InitialSimplex(best, g_dim, g_amount * 0.33)
best = g_simplex[0][:]

for restarts in range(99999):
  for ii in range(g_dim * 5):
    g_simplex.sort()
    print("reflect", ii, g_simplex[0])
    Reflect(g_simplex, g_binary)

  mulli = 0.1 + 15 * random.random()**2.0
  g_codecs = RandomizedJxlCodecs()
  g_best_val = None # get a new best val
  print("\n\n\nRestart", restarts, "mulli", mulli)
  g_simplex.sort()
  best = g_simplex[0][:]
  g_simplex = InitialSimplex(best, g_dim, g_amount * mulli)



#!/usr/bin/env python3

# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""Produces demos for how progressive-saliency encoding would look like.

As long as we do not have a progressive decoder that allows showing images
generated from partially-available data, we can resort to building
animated gifs that show how progressive loading would look like.

Method:

1. JPEG-XL encode the image, but stop at the pre-final (2nd) step.
2. Use separate tool to compute a heatmap which shows where differences between
   the pre-final and final image are expected to be perceptually worst.
3. Use this heatmap to JPEG-XL encode the image with the final step split into
   'salient parts only' and 'non-salient parts'. Generate a sequence of images
   that stop decoding after the 1st, 2nd, 3rd, 4th step. JPEG-XL decode these
   truncated images back to PNG.
4. Measure byte sizes of the truncated-encoded images.
5. Build an animated GIF with variable delays by calling ImageMagick's
   `convert` command.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from six.moves import zip
import ast  # For ast.literal_eval() only.
import os
import re
import shlex
import subprocess
import sys

_BLOCKSIZE = 8

_CONF_PARSERS = dict(
    keep_tempfiles=lambda s: bool(ast.literal_eval(s)),
    heatmap_command=shlex.split,
    simulated_progressive_loading_time_sec=float,
    simulated_progressive_loading_delay_until_looparound_sec=float,
    jpegxl_encoder=shlex.split,
    jpegxl_decoder=shlex.split,
    blurring=lambda s: s.split(),
)


def parse_config(config_filename):
  """Parses the configuration file."""
  conf = {}
  re_comment = re.compile(r'^\s*(?:#.*)?$')
  re_param = re.compile(r'^(?P<option>\w+)\s*:\s*(?P<value>.*?)\s*$')
  try:
    with open(config_filename) as h:
      for line in h:
        if re_comment.match(line):
          continue
        m = re_param.match(line)
        if not m:
          raise ValueError('Syntax error')
        conf[m.group('option')] = (
            _CONF_PARSERS[m.group('option')](m.group('value')))
  except Exception as exn:
    raise ValueError('Bad Configuration line ({}): {}'.format(exn, line))
  missing_options = set(_CONF_PARSERS) - set(conf)
  if missing_options:
    raise ValueError('Missing configuration options: ' + ', '.join(
        sorted(missing_options)))
  return conf


def generate_demo_image(config, input_filename, output_filename):
  tempfiles = []
  #
  def encode_img(input_filename, output_filename, num_steps,
                 heatmap_filename=None):
    replacements = {
        '${INPUT}': input_filename,
        '${OUTPUT}': output_filename,
        '${STEPS}': str(num_steps),
        # Heatmap argument will be provided in --param=value form.
        '${HEATMAP_ARG}': ('--saliency_map_filename=' + heatmap_filename
                           if heatmap_filename is not None else '')
        }
    # Remove empty args. This removes the heatmap-argument if no heatmap
    # is provided..
    cmd = [
        _f for _f in
        [replacements.get(arg, arg) for arg in config['jpegxl_encoder']] if _f
    ]
    tempfiles.append(output_filename)
    subprocess.call(cmd)
  #
  def decode_img(input_filename, output_filename):
    replacements = {'${INPUT}': input_filename, '${OUTPUT}': output_filename}
    cmd = [replacements.get(arg, arg) for arg in config['jpegxl_decoder']]
    tempfiles.append(output_filename)
    subprocess.call(cmd)
  #
  def generate_heatmap(orig_image_filename, coarse_grained_filename,
                       heatmap_filename):
    cmd = config['heatmap_command'] + [
        str(_BLOCKSIZE), orig_image_filename, coarse_grained_filename,
        heatmap_filename]
    tempfiles.append(heatmap_filename)
    subprocess.call(cmd)
  #
  try:
    encode_img(input_filename, output_filename + '._step1.pik', 1)
    decode_img(output_filename + '._step1.pik', output_filename + '._step1.png')
    encode_img(input_filename, output_filename + '._step2.pik', 2)
    decode_img(output_filename + '._step2.pik', output_filename + '._step2.png')
    generate_heatmap(input_filename, output_filename + '._step2.png',
                     output_filename + '._heatmap.png')
    encode_img(input_filename,
               output_filename + '._step3.pik', 3,
               output_filename + '._heatmap.png')
    encode_img(input_filename,
               output_filename + '._step4.pik', 4,
               output_filename + '._heatmap.png')
    decode_img(output_filename + '._step3.pik', output_filename + '._step3.png')
    decode_img(output_filename + '._step4.pik', output_filename + '._step4.png')
    data_sizes = [
        os.stat('{}._step{}.pik'.format(output_filename, num_step)).st_size
        for num_step in (1, 2, 3, 4)]
    time_offsets = [0] + [
        # Imagemagick's `convert` accepts delays in units of 1/100 sec.
        round(100 * config['simulated_progressive_loading_time_sec'] * size /
              data_sizes[-1]) for size in data_sizes]
    time_delays = [t_next - t_prev
                   for t_next, t_prev in zip(time_offsets[1:], time_offsets)]
    # Add a fake white initial image. As long as no usable image data is
    # available, the user will see a white background.
    subprocess.call(['convert',
                     output_filename + '._step1.png',
                     '-fill', 'white', '-colorize', '100%',
                     output_filename + '._step0.png'])
    tempfiles.append(output_filename + '._step0.png')
    subprocess.call(
        ['convert', '-loop', '0', output_filename + '._step0.png'] +
        [arg for args in [
            ['-delay', str(time_delays[n - 1]),
             '-blur', config['blurring'][n - 1],
             '{}._step{}.png'.format(output_filename, n)]
            for n in (1, 2, 3, 4)] for arg in args] +
        ['-delay', str(round(100 * config[
            'simulated_progressive_loading_delay_until_looparound_sec'])),
         output_filename + '._step4.png',
         output_filename])
  finally:
    if not config['keep_tempfiles']:
      for filename in tempfiles:
        try:
          os.unlink(filename)
        except OSError:
          pass  # May already have been deleted otherwise.


def main():
  if sys.version.startswith('2.'):
    sys.exit('This is a python3-only script.')
  if (len(sys.argv) != 4 or not sys.argv[-1].endswith('.gif')
      or not sys.argv[-2].endswith('.png')):
    sys.exit(
        'Usage: {} [config_options_file] [input.png] [output.gif]'.format(
            sys.argv[0]))
  try:
    _, config_filename, input_filename, output_filename = sys.argv
    config = parse_config(config_filename)
    generate_demo_image(config, input_filename, output_filename)
  except ValueError as exn:
    sys.exit(exn)



if __name__ == '__main__':
  main()



#!/usr/bin/env python3
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import sys

def find_key(entries : list[str], key: str) -> int:
  prefix = f"{key.lower()}: "
  for i in range(len(entries)):
    if entries[i].lower().startswith(prefix):
      return i
  return -1

def set_value(entries: list[str], key: str, value: str):
  new_line = f'{key}: {value}'
  # TODO(eustas): deal with repeated items
  idx = find_key(entries, key)
  if idx < 0:
    entries.append(new_line)
  else:
    entries[idx] = new_line

def transform_deb_822(archs):
  sources_path = "/etc/apt/sources.list.d/debian.sources"
  with open(sources_path) as f:
    lines = [line.rstrip() for line in f]
  lines.append('')
  entries = []
  entry = []
  for line in lines:
    if len(line) == 0:
      if len(entry) > 0:
        entries.append(entry)
      entry = []
    else:
      entry.append(line)

  new_entries = []
  for entry in entries:
    types_key = find_key(entry, "Types")
    if types_key < 0:
      continue
    if "types: deb" != entry[types_key].lower():
      continue
    deb_entry = entry[:]
    for arch in archs:
      deb_entry.append(f"Architectures-Add: {arch}")
    new_entries.append(deb_entry)
    deb_src_entry = deb_entry[:]
    set_value(deb_src_entry, "Types", "deb-src")
    new_entries.append(deb_src_entry)

  new_lines = []
  for entry in new_entries:
    if len(new_lines) > 0:
      new_lines.append("")
    new_lines.extend(entry)

  with open(sources_path, "w") as f:
    f.write('\n'.join(new_lines))

def main():
  if len(sys.argv) != 2:
    print(f"Usage: {sys.argv[1]} ARCHS")
    sys.exit(1)
  archs_str = sys.argv[1]
  archs = archs_str.split(',')
  if True:
    transform_deb_822(archs)
  else:
    sys.exit(1)

if __name__ == '__main__':
  main()



#!/usr/bin/python
# Copyright (c) the JPEG XL Project Authors. All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""apply_simplex.py: Updates constants based on results of simplex search.

To use this tool, the simplex search parameters must we wrapped in a bias(n)
function call that returns the value of the VARn environment variable. The
tool reads a text file containing the simplex definition that simplex_fork.py
has written, and updates the target source files by substituting the bias(n)
function calls with the (n+1)th coordinate of the simplex vector, and also
simplifies these expressions by evaluating them to a single floating point
literal.

The tool recognizes and evaluates the following expressions:
  <constant> + bias(n),
  <constant> * bias(n),
  <constant> + <coeff> * bias(n).

The --keep_bias command-line flag can be used to continue an aborted simplex
search. This will keep the same bias(n) terms in the code, but would update the
surrounding constants.

The --index_min and --index_max flags can be used to update only a subset of the
bias(n) parameters.
"""

import argparse
import re
import sys

def ParseSimplex(fn):
    """Returns the simplex definition written by simplex_fork.py"""

    with open(fn, "r") as f:
        line = f.readline()
        vec = eval(line)
    return vec


def PythonExpr(c_expr):
    """Removes the f at the end of float literals"""

    def repl(m):
        return m.group(1)

    return re.sub("(\d+)f", repl, c_expr)


def UpdateSourceFile(fn, vec, keep_bias, id_min, id_max, minval):
    """Updates expressions containing a bias(N) term."""

    with open(fn, "r") as f:
        lines_in = f.readlines()
        lines_out = []
        rbias = "(bias\((\d+)\))"
        r = " -?\d+\.\d+f?( (\+|-|\*) (\d+\.\d+f? \* )?" + rbias + ")"
        for line in lines_in:
            line_out = line
            x = re.search(r, line)
            if x:
                id = int(x.group(5))
                if id >= id_min and id <= id_max:
                    expr = re.sub(rbias, str(vec[id + 1]), x.group(0))
                    val = eval(PythonExpr(expr))
                    if minval and val < minval:
                        val = minval
                    expr_out = " " + str(val) + "f"
                    if keep_bias:
                        expr_out += x.group(1)
                    line_out = re.sub(r, expr_out, line)
            lines_out.append(line_out)

    with open(fn, "w") as f:
        f.writelines(lines_out)
        f.close()


def ApplySimplex(args):
  """Main entry point of the program after parsing parameters."""

  vec = ParseSimplex(args.simplex)
  for fn in args.target:
      UpdateSourceFile(fn, vec, args.keep_bias, args.index_min, args.index_max,
                       args.minval)
  return 0


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument('target', type=str, nargs='+',
                      help='source file(s) to update')
  parser.add_argument('--simplex', default='best_simplex.txt',
                      help='simplex to apply to the code')
  parser.add_argument('--keep_bias', default=False, action='store_true',
                      help='keep the bias term in the code, can be used to ' +
                      'continue simplex search')
  parser.add_argument('--index_min', type=int, default=0,
                      help='start index of the simplex to apply')
  parser.add_argument('--index_max', type=int, default=9999,
                      help='last index of the simplex to apply')
  parser.add_argument('--minval', type=float, default=None,
                      help='apply a minimum to expression results')
  args = parser.parse_args()
  sys.exit(ApplySimplex(args))


if __name__ == '__main__':
  main()
